# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Fixtures for tests"""
import pytest

from pybest.geminals import ROOpCCD
from pybest.wrappers.hf import RHF
from pybest.wrappers.tests.common import Molecule

############################>Dipole moment<############################
#
# Define fixtures for testing: water and helium
# All in debye unit
# [ (molecule name), (expected) ]
test_hf_dm = [
    (
        "h2o",
        {
            "dipole_hf": [0.04129661, 2.25353156, 0.00000000],
        },
    ),
    (
        "water",
        {"dipole_hf": [0.00000000, -1.98081762, 0.00000000]},
    ),
    (
        "he",
        {"dipole_hf": [0.00000000, 0.00000000, 0.00000000]},
    ),
]

test_opccd_dm = [
    (
        "h2o",
        ROOpCCD,
        {
            "dipole_pccd": [0.2797343, 1.9267571, 0.00000000],
            "dipole_lccsd": [-0.01807735, 2.20762975, 0.00000000],
        },
    ),
    (
        "he",
        ROOpCCD,
        {
            "dipole_pccd": [0.00000000, 0.00000000, 0.00000000],
            "dipole_lccsd": [0.00000000, 0.00000000, 0.00000000],
        },
    ),
]


@pytest.fixture(params=test_hf_dm)
def mol_generation(request, linalg):
    """Generate molecule for HF calculations and multipole moments"""
    mol = Molecule(linalg, "cc-pvdz", f"test/{request.param[0]}.xyz", alpha=0)
    mol.do_scf(RHF)
    mol.compute_dipole()

    # Return molecule and expected values
    return mol, request.param[1]


@pytest.fixture(params=test_opccd_dm)
def mol_generation_oopccd(request, linalg):
    """Generate molecule for pCCD-based calculations"""
    mol = Molecule(linalg, "cc-pvdz", f"test/{request.param[0]}.xyz", alpha=0)
    mol.do_scf(RHF)
    mol.do_pccd(request.param[1])
    mol.compute_dipole()

    # Return molecule and expected values
    return mol, request.param[2]


############################>Quadrupole moment<############################
test_hf_quadrupole_moment = [
    (
        "water",
        {
            "quadrupole_hf": [
                -1.58341900,  # qm_xx
                0.09143664,  # qm_yy
                1.49198236,  # qm_zz
                0.00000000,  # qm_xy
                0.00000000,  # qm_xz
                0.00000000,  # qm_yz
            ],
        },
    ),
]

test_opccd_quadrupole_moment = [
    (
        "he",
        ROOpCCD,
        {
            "quadrupole_pccd": [
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
            ],
            "quadrupole_lccsd": [
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
            ],
        },
    ),
]


@pytest.fixture(params=test_hf_quadrupole_moment)
def mol_generation_quadrupole_moment(request, linalg):
    """Generate molecule for the HF quadrupole moment calculation"""
    mol = Molecule(linalg, "cc-pvdz", f"test/{request.param[0]}.xyz", alpha=0)
    mol.do_scf(RHF)
    mol.compute_quadrupole()

    # Return molecule and expected values
    return mol, request.param[1]


@pytest.fixture(params=test_opccd_quadrupole_moment)
def mol_generation_oopccd_qm(request, linalg):
    """Generate molecule for pCCD-based calculations"""
    mol = Molecule(linalg, "cc-pvdz", f"test/{request.param[0]}.xyz", alpha=0)
    mol.do_scf(RHF)
    mol.do_pccd(request.param[1])
    mol.compute_quadrupole()

    # Return molecule and expected values
    return mol, request.param[2]
